import io
import pathlib
from copy import deepcopy
from typing import List, Literal, Optional, Tuple

import ee
import numpy as np
import pandas as pd
import rasterio as rio
import utm


def query_utm_crs_info(lon: float, lat: float) -> Tuple[float, float, str]:
    """Converts a pair of lat, lon to UTM coordinates

    Args:
        lon (float): The longitude of the point
        lat (float): The latitude of the point
    Returns:
        Tuple[float, float, str]: The UTM coordinates and the
            EPSG code of the zone
    """
    x, y, zone, _ = utm.from_latlon(lat, lon)
    zone_epsg = f"326{zone:02d}" if lat >= 0 else f"327{zone:02d}"
    return x, y, "EPSG:" + zone_epsg


def quadsplit_manifest(manifest: dict) -> List[dict]:
    """Splits a manifest in 4 smaller manifests

    Args:
        manifest (dict): The manifest to be split

    Returns:
        List[dict]: A list of 4 manifests
    """

    # deep copy the manifest
    manifest_copy = deepcopy(manifest)
    new_width = manifest["grid"]["dimensions"]["width"] // 2
    new_height = manifest["grid"]["dimensions"]["height"] // 2
    manifest_copy["grid"]["dimensions"]["width"] = new_width
    manifest_copy["grid"]["dimensions"]["height"] = new_height

    manifests = []
    for idx in range(4):

        # Load a new manifest
        new_manifest = deepcopy(manifest_copy)

        # Set the scale
        res_x = manifest["grid"]["affineTransform"]["scaleX"]
        res_y = manifest["grid"]["affineTransform"]["scaleY"]

        # Adjust the width and height
        if idx == 0:
            add_x = 0
            add_y = 0
        if idx == 1:
            add_x = new_width * res_x
            add_y = 0
        elif idx == 2:
            add_x = 0
            add_y = new_height * res_y
        elif idx == 3:
            add_x = new_width * res_x
            add_y = new_height * res_y

        # Adjust the translation
        new_manifest["grid"]["affineTransform"]["translateX"] += add_x
        new_manifest["grid"]["affineTransform"]["translateY"] += add_y

        # Append the new manifest
        manifests.append(new_manifest)

    return manifests


def computePixels_np(
    manifest_dict: dict,
    quiet: Optional[bool] = False,
    deep_level: Optional[int] = 0,
) -> np.ndarray:
    """Implements the computePixels method from
    the Earth Engine API. If the image is too large,
    it splits the image in 4 and downloads the data
    in batches.

    Args:
        manifest_dict (dict): The manifest to be downloaded

    Returns:
        np.ndarray: The image as a numpy array
    """
    if deep_level == 5:
        raise ValueError(
            "Deep level reached 5. Aborting." f" Manifest: {manifest_dict}"
        )

    try:
        # Download the data
        data = ee.data.computePixels(manifest_dict)
        with rio.open(io.BytesIO(data)) as f:
            data_np = f.read()
    except Exception as e:

        # Create a container for the data
        data_np = np.zeros(
            (
                len(manifest_dict["bandIds"]),
                manifest_dict["grid"]["dimensions"]["width"],
                manifest_dict["grid"]["dimensions"]["height"],
            )
        )

        # Split the manifest in 4
        manifest_dicts = quadsplit_manifest(manifest_dict)

        for idx, manifest_dict_batch in enumerate(manifest_dicts):
            if not quiet:
                print(f"Downloading {idx} of 4...")

            # Try to obtain the data for the batch
            dnp = computePixels_np(
                manifest_dict=manifest_dict_batch,
                quiet=quiet,
                deep_level=deep_level + 1,
            )

            # Insert the data in the container
            if idx == 0:
                data_np[:, : dnp.shape[1], : dnp.shape[2]] = dnp
            elif idx == 1:
                data_np[:, : dnp.shape[1], -dnp.shape[2] :] = dnp
            elif idx == 2:
                data_np[:, -dnp.shape[1] :, : dnp.shape[2]] = dnp
            elif idx == 3:
                data_np[:, -dnp.shape[1] :, -dnp.shape[2] :] = dnp

    return data_np


def getPixels_np(
    manifest_dict: dict,
    quiet: Optional[bool] = False,
    deep_level: Optional[int] = 0,
) -> np.ndarray:
    """Implements the getPixels method from
    the Earth Engine API. If the image is too large,
    it splits the image in 4 and downloads the data
    in batches.

    Args:
        manifest_dict (dict): The manifest to be downloaded

    Returns:
        np.ndarray: The image as a numpy array
    """
    if deep_level == 5:
        raise ValueError(
            "Deep level reached 5. Aborting." f" Manifest: {manifest_dict}"
        )

    try:
        # Download the data
        data = ee.data.getPixels(manifest_dict)
        with rio.open(io.BytesIO(data)) as f:
            data_np = f.read()
    except Exception as e:
        # Create a container for the data
        data_np = np.zeros(
            (
                len(manifest_dict["bandIds"]),
                manifest_dict["grid"]["dimensions"]["width"],
                manifest_dict["grid"]["dimensions"]["height"],
            )
        )

        # Split the manifest in 4
        manifest_dicts = quadsplit_manifest(manifest_dict)

        for idx, manifest_dict_batch in enumerate(manifest_dicts):
            if not quiet:
                print(f"Downloading {idx} of 4...")

            # Obtain the data for the batch
            dnp = getPixels_np(
                manifest_dict=manifest_dict_batch,
                quiet=quiet,
                deep_level=deep_level + 1,
            )

            # Insert the data in the container
            if idx == 0:
                data_np[:, : dnp.shape[1], : dnp.shape[2]] = dnp
            elif idx == 1:
                data_np[:, : dnp.shape[1], -dnp.shape[2] :] = dnp
            elif idx == 2:
                data_np[:, -dnp.shape[1] :, : dnp.shape[2]] = dnp
            elif idx == 3:
                data_np[:, -dnp.shape[1] :, -dnp.shape[2] :] = dnp

    return data_np


def getImage_batch(
    row: pd.Series,
    output_path: str,
    type: Literal["getPixels", "computePixels"],
    quiet: Optional[bool] = False,
) -> pathlib.Path:
    """Downloads the image from the manifest as a
    GeoTIFF file.

    Args:
        manifest_dict (dict): The manifest to be downloaded
        output_path (str): The path where the file will be saved
    """
    if not quiet:
        print(f"Downloading {row.outname}...")

    # Load the manifest
    manifest_dict = eval(row.manifest)

    # Download the data
    if type == "computePixels":
        manifest_dict["expression"] = ee.deserializer.decode(
            eval(manifest_dict["expression"])
        )
        data_np = computePixels_np(manifest_dict, quiet=quiet)
    elif type == "getPixels":
        data_np = getPixels_np(manifest_dict, quiet=quiet)
    else:
        raise ValueError("type must be either 'getPixels' or 'computePixels'")

    # Save the data
    metadata_rio = {
        "driver": "GTiff",
        "count": data_np.shape[0],
        "dtype": data_np.dtype,
        "height": manifest_dict["grid"]["dimensions"]["height"],
        "width": manifest_dict["grid"]["dimensions"]["width"],
        "transform": rio.Affine(
            manifest_dict["grid"]["affineTransform"]["scaleX"],
            manifest_dict["grid"]["affineTransform"]["shearX"],
            manifest_dict["grid"]["affineTransform"]["translateX"],
            manifest_dict["grid"]["affineTransform"]["shearY"],
            manifest_dict["grid"]["affineTransform"]["scaleY"],
            manifest_dict["grid"]["affineTransform"]["translateY"],
        ),
        "crs": manifest_dict["grid"]["crsCode"],
    }

    # Create the output folder
    outfile = pathlib.Path(output_path) / (row.outname)
    outfile.parent.mkdir(parents=True, exist_ok=True)

    # Save the data
    with rio.open(outfile, "w", **metadata_rio) as dst:
        dst.write(data_np)

    return outfile



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastcubo']

package_data = \
{'': ['*']}

install_requires = \
['pandas>=2.0.0', 'utm>=0.7.0,<0.8.0']

setup_kwargs = {
    'name': 'fastcubo',
    'version': '0.0.99',
    'description': '🏎💨vroom vroom - data downloader',
    'long_description': '# FastCubo\n\nA simple API for `ee.data.pixels` inspired by [cubo](https://github.com/ESDS-Leipzig/cubo), designed for creating and managing data cubes up to 10 times faster.\n\n## Installation\n\nInstall the latest version from PyPI:\n\n```bash\npip install fastcubo\n```\n\n## How to use\n\n\nDownload a ee.Image\n\n```python\nimport ee\nimport fastcubo\n\nee.Initialize(opt_url="https://earthengine-highvolume.googleapis.com")\n\n\ntable = fastcubo.query_getPixels_image(\n    points=[(-76.5, -9.5), (-76.5, -10.5), (-77.5, -10.5)],\n    collection="NASA/NASADEM_HGT/001",\n    bands=["elevation"],\n    edge_size=128,\n    resolution=90\n)\n\nfastcubo.getPixels(table, nworkers=4, output_path="demo1")\n```\n\nDownload a ee.ImageCollection\n\n```python\nimport fastcubo\nimport ee\n\nee.Initialize(opt_url="https://earthengine-highvolume.googleapis.com")\n\ntable = fastcubo.query_getPixels_imagecollection(\n    point=(51.079225, 10.452173),\n    collection="COPERNICUS/S2_HARMONIZED", # Id of the GEE collection\n    bands=["B4","B3","B2"], # Bands to retrieve\n    data_range=["2016-06-01", "2017-07-01"], # Date range of the data\n    edge_size=128, # Edge size of the cube (px)\n    resolution=10, # Pixel size of the cube (m)\n)\nfastcubo.getPixels(table, nworkers=4, output_path="demo2")\n```\n\n\nDownload a ee.Image Compute Pixels\n\n```python\nimport fastcubo\nimport ee\n\nee.Initialize(opt_url="https://earthengine-highvolume.googleapis.com")\n\ntable = fastcubo.query_computePixels_image(\n    points=[(-76.5, -9.5), (-76.5, -10.5), (-77.5, -10.5)],\n    expression=ee.Image("NASA/NASADEM_HGT/001").divide(1000),\n    bands=["elevation"],\n    edge_size=128,\n    resolution=90\n)\nfastcubo.computePixels(table, nworkers=4, output_path="demo3")\n```',
    'author': 'Cesar Aybar',
    'author_email': 'fcesar.aybar@uv.es',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/csaybar/fastcubo',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import concurrent.futures
import pathlib
import re
from typing import List, Optional, Tuple, Union

import ee
import pandas as pd

from fastcubo.utils import getImage_batch, query_utm_crs_info


def query_getPixels_image(
    points: List[Tuple[float, float]],
    collection: str,
    bands: List[str],
    edge_size: float,
    resolution: float,
    outnames: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Returns a DataFrame with the metadata needed to
    retrieve the data using `ee.data.getPixels`

    Args:
        points (List[Tuple[float, float]]): The centroid
            of the square to be queried.
        outnames (List[str]): The name of the output files
            to be saved.
        collection (str): The collection to be queried
        bands (List[str]): The bands to be queried
        start_date (str): The start date of the query
        end_date (str): The end date of the query
        edge_size (float): The size of the square to be queried
        resolution (float): The resolution of the query

    Returns:
        pd.DataFrame: A DataFrame with the metadata needed to
            retrieve the data using `ee.data.getPixels`
    """

    if outnames is None:
        basename = collection.replace("/", "_")
        outnames = [f"%s__%04d.tif" % (basename, i) for i in range(len(points))]

    # From EPSG to UTM
    epsg = [query_utm_crs_info(lon, lat) for lon, lat in points]
    lon_utm, lat_utm, zone_epsg = zip(*epsg)

    # Fix the center of the square to be the centroid
    lon_utm = [x - edge_size * resolution / 2 for x in lon_utm]
    lat_utm = [y + edge_size * resolution / 2 for y in lat_utm]

    # Create the query_table
    query_table = pd.DataFrame(
        {
            "lon": [lon for lon, _ in points],
            "lat": [lat for _, lat in points],
            "x": lon_utm,
            "y": lat_utm,
            "epsg": zone_epsg,
            "collection": collection,
            "bands": ", ".join(bands),
            "edge_size": edge_size,
            "resolution": resolution,
        }
    )

    # Add manifest to the query_table
    manifests = []
    for index, row in query_table.iterrows():
        manifest = {
            "assetId": row["collection"],
            "fileFormat": "GEO_TIFF",
            "bandIds": bands,
            "grid": {
                "dimensions": {
                    "width": row["edge_size"],
                    "height": row["edge_size"],
                },
                "affineTransform": {
                    "scaleX": row["resolution"],
                    "shearX": 0,
                    "translateX": row["x"],
                    "shearY": 0,
                    "scaleY": -row["resolution"],
                    "translateY": row["y"],
                },
                "crsCode": zone_epsg[index],
            },
        }
        manifests.append(str(manifest))
    query_table["manifest"] = manifests

    # Save the outnames
    query_table["outname"] = outnames

    return query_table


def query_getPixels_imagecollection(
    point: Tuple[float, float],
    collection: str,
    bands: List[str],
    edge_size: float,
    resolution: float,
    data_range: Tuple[str, str],
    outnames: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Returns a DataFrame with the metadata needed to
    retrieve the data using `ee.data.getPixels`

    Args:
        task_id (str): The id of the task to be executed
        points (List[Tuple[float, float]]): The centroid
            of the square to be queried.
        outnames (List[str]): The name of the output files
            to be saved.
        collection (str): The collection to be queried
        bands (List[str]): The bands to be queried
        data_range (Tuple[str, str]): The range of dates to
            be queried in the format (start_date, end_date).
        edge_size (float): The size of the square to be queried
        resolution (float): The resolution of the query

    Returns:
        pd.DataFrame: A DataFrame with the metadata needed to
            retrieve the data using `ee.data.getPixels`
    """
    # From EPSG to UTM
    lon_utm, lat_utm, zone_epsg = query_utm_crs_info(*point)

    # Fix the center of the square to be the centroid
    lon_utm = lon_utm - edge_size * resolution / 2
    lat_utm = lat_utm + edge_size * resolution / 2

    # Get the images
    images = (
        ee.ImageCollection(collection)
        .filterBounds(ee.Geometry.Point((lon_utm, lat_utm), proj=zone_epsg))
        .filterDate(data_range[0], data_range[1])
        .select(bands)
    )

    # Get the ids and dates
    ids = images.aggregate_array("system:id").getInfo()
    dates = images.aggregate_array("system:time_start").getInfo()
    dates_str = [
        pd.to_datetime(date, unit="ms").strftime("%Y-%m-%d %H:%M:%S") for date in dates
    ]

    # Create the query_table
    query_table = pd.DataFrame(
        {
            "lon": point[0],
            "lat": point[1],
            "x": lon_utm,
            "y": lat_utm,
            "epsg": zone_epsg,
            "collection": collection,
            "bands": ", ".join(bands),
            "edge_size": edge_size,
            "resolution": resolution,
            "img_id": ids,
            "img_date": dates_str,
        }
    )

    # Add manifest to the query_table
    manifests = []
    for index, row in query_table.iterrows():
        manifest = {
            "assetId": row["img_id"],
            "fileFormat": "GEO_TIFF",
            "bandIds": bands,
            "grid": {
                "dimensions": {
                    "width": row["edge_size"],
                    "height": row["edge_size"],
                },
                "affineTransform": {
                    "scaleX": row["resolution"],
                    "shearX": 0,
                    "translateX": row["x"],
                    "shearY": 0,
                    "scaleY": -row["resolution"],
                    "translateY": row["y"],
                },
                "crsCode": row["epsg"],
            },
        }
        manifests.append(str(manifest))

    # Save the manifests
    query_table["manifest"] = manifests

    if outnames is None:
        query_table["outname"] = query_table["img_id"].apply(
            lambda x: pathlib.Path(x).stem + ".tif"
        )
    else:
        query_table["outname"] = outnames

    return query_table


def getPixels(
    table: pd.DataFrame,
    nworkers: Optional[int] = None,
    output_path: Union[str, pathlib.Path, None] = None,
    quiet: bool = False,
) -> None:
    """Create a GeoTIFF file from a query_table

    Args:
        qtable (pd.DataFrame): The query_table to be downloaded
        output_path (Optional[str], optional): The path to save
            the file. Defaults to None.
    """

    # Save the output_path
    if output_path is None:
        output_path = pathlib.Path(table.task_id[0])
        output_path.mkdir(parents=True, exist_ok=True)
    else:
        output_path = pathlib.Path(output_path)

    # Using ThreadPoolExecutor to manage concurrent downloads
    with concurrent.futures.ThreadPoolExecutor(max_workers=nworkers) as executor:
        futures = [
            executor.submit(getImage_batch, row, output_path, "getPixels", quiet)
            for _, row in table.iterrows()
        ]
        # If there is an output_path, raise it
        results = []
        for future in concurrent.futures.as_completed(futures):
            if future.exception() is not None:
                raise future.exception()
            results.append(future.result())

    return results


def query_computePixels_image(
    points: List[Tuple[float, float]],
    expression: ee.Image,
    bands: List[str],
    edge_size: float,
    resolution: float,
    outnames: Optional[List[str]] = None,
) -> pd.DataFrame:
    """Returns a DataFrame with the metadata needed to
    retrieve the data using `ee.data.getPixels`

    Args:
        points (List[Tuple[float, float]]): The centroid
            of the square to be queried.
        outnames (List[str]): The name of the output files
            to be saved.
        collection (str): The collection to be queried
        bands (List[str]): The bands to be queried
        start_date (str): The start date of the query
        end_date (str): The end date of the query
        edge_size (float): The size of the square to be queried
        resolution (float): The resolution of the query

    Returns:
        pd.DataFrame: A DataFrame with the metadata needed to
            retrieve the data using `ee.data.getPixels`
    """
    if outnames is None:
        regex_exp = re.compile(r'"constantValue":\s*"([^"]*)"')
        product_id = re.findall(regex_exp, expression.serialize())[0]
        basename = product_id.replace("/", "_")
        outnames = [f"%s__%04d.tif" % (basename, i) for i in range(len(points))]

    # From EPSG to UTM
    epsg = [query_utm_crs_info(lon, lat) for lon, lat in points]
    lon_utm, lat_utm, zone_epsg = zip(*epsg)

    # Fix the center of the square to be the centroid
    lon_utm = [x - edge_size * resolution / 2 for x in lon_utm]
    lat_utm = [y + edge_size * resolution / 2 for y in lat_utm]

    # Create the query_table
    query_table = pd.DataFrame(
        {
            "lon": [lon for lon, _ in points],
            "lat": [lat for _, lat in points],
            "x": lon_utm,
            "y": lat_utm,
            "epsg": zone_epsg,
            "expression": expression.serialize(),
            "bands": ", ".join(bands),
            "edge_size": edge_size,
            "resolution": resolution,
        }
    )

    # Add manifest to the query_table
    manifests = []
    for index, row in query_table.iterrows():
        manifest = {
            "expression": row["expression"],
            "fileFormat": "GEO_TIFF",
            "bandIds": bands,
            "grid": {
                "dimensions": {
                    "width": row["edge_size"],
                    "height": row["edge_size"],
                },
                "affineTransform": {
                    "scaleX": row["resolution"],
                    "shearX": 0,
                    "translateX": row["x"],
                    "shearY": 0,
                    "scaleY": -row["resolution"],
                    "translateY": row["y"],
                },
                "crsCode": zone_epsg[index],
            },
        }
        manifests.append(str(manifest))
    query_table["manifest"] = manifests

    # Save the outnames
    query_table["outname"] = outnames

    return query_table


def computePixels(
    table: pd.DataFrame,
    nworkers: Optional[int] = None,
    output_path: Union[str, pathlib.Path, None] = None,
    quiet: bool = False,
) -> None:
    """Create a GeoTIFF file from a query_table

    Args:
        qtable (pd.DataFrame): The query_table to be downloaded
        output_path (Optional[str], optional): The path to save
            the file. Defaults to None.
    """

    # Save the output_path
    if output_path is None:
        output_path = pathlib.Path(".")
    else:
        output_path = pathlib.Path(output_path)

    # Using ThreadPoolExecutor to manage concurrent downloads
    with concurrent.futures.ThreadPoolExecutor(max_workers=nworkers) as executor:
        futures = [
            executor.submit(getImage_batch, row, output_path, "computePixels", quiet)
            for _, row in table.iterrows()
        ]
        # If there is an output_path, raise it
        results = []
        for future in concurrent.futures.as_completed(futures):
            if future.exception() is not None:
                raise future.exception()
            results.append(future.result())

    return results
